/********************************************
* EC Flashr 1.0
* (c) Tammo Winkler
* 2025
* Flasher Tool for BU1 Firmware
********************************************/
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

int zaehleZeilen(FILE *datei) {
  char zeile[5000];
  int anzahl = 0;
  while (fgets(zeile, sizeof(zeile), datei) != NULL) {
    int i = 0;
    while (zeile[i] != '\0' && (zeile[i] == ' ' || zeile[i] == '\t')) {
      i++;
    }
    if (zeile[i] != '\0' &&
        (isdigit(zeile[i]) || (zeile[i] >= 'A' && zeile[i] <= 'F') ||
         (zeile[i] >= 'a' && zeile[i] <= 'f'))) {
      anzahl++;
    }
  }
  return anzahl;
}
int open_serial_port(const char *device, uint32_t baud_rate) {
  int fd = open(device, O_RDWR | O_NOCTTY);
  if (fd == -1) {
    perror(device);
    return -1;
  }
  int result = tcflush(fd, TCIOFLUSH);
  if (result) {
    perror("tcflush failed");
  }
  struct termios options;
  result = tcgetattr(fd, &options);
  if (result) {
    perror("tcgetattr failed");
    close(fd);
    return -1;
  }
  options.c_cflag &= ~PARENB;
  options.c_cflag &= ~CSTOPB;
  options.c_cflag &= ~CSIZE;
  options.c_cflag |= CS8;
  options.c_cflag &= ~CRTSCTS;
  options.c_cflag |= CREAD | CLOCAL;
  options.c_lflag &= ~ICANON;
  options.c_lflag &= ~ECHO;
  options.c_lflag &= ~ECHOE;
  options.c_lflag &= ~ECHONL;
  options.c_lflag &= ~ISIG;
  options.c_iflag &= ~(IXON | IXOFF | IXANY);
  options.c_iflag &=
      ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL);
  options.c_oflag &= ~OPOST;
  options.c_oflag &= ~ONLCR;
  options.c_cc[VTIME] = 10;
  options.c_cc[VMIN] = 0;
  switch (baud_rate) {
  case 4800:
    cfsetospeed(&options, B4800);
    break;
  case 9600:
    cfsetospeed(&options, B9600);
    break;
  case 19200:
    cfsetospeed(&options, B19200);
    break;
  case 38400:
    cfsetospeed(&options, B38400);
    break;
  case 115200:
    cfsetospeed(&options, B115200);
    break;
  default:
    fprintf(stderr, "warning: baud rate %u is not supported, using 9600.\n",
            baud_rate);
    cfsetospeed(&options, B9600);
    break;
  }
  cfsetispeed(&options, cfgetospeed(&options));
  result = tcsetattr(fd, TCSANOW, &options);
  if (result) {
    perror("tcsetattr failed");
    close(fd);
    return -1;
  }
  sleep(2);
  tcflush(fd, TCIOFLUSH);
  return fd;
}
int write_port(int fd, uint8_t *buffer, size_t size) {
  ssize_t result = write(fd, buffer, size);
  if (result != (ssize_t)size) {
    perror("failed to write to port");
    return -1;
  }
  return size;
}
ssize_t read_port(int fd, uint8_t *buffer, size_t size) {
  size_t received = 0;
  while (received < size) {
    ssize_t r = read(fd, buffer + received, size - received);
    if (r < 0) {
      perror("failed to read from port");
      return -1;
    }
    if (r == 0) {
      break;
    }
    received += r;
  }
  return received;
}
int main(int argc, char *argv[]) {
  printf("\nFlashr\nEngineering-Connection Flasher Tool\n\n");
  if (argc == 3) {
    printf("Serial Device is %s\n", argv[1]);
    printf("Firmware File is %s\n", argv[2]);
  } else {
    printf("Usage: flashr <serial device> <firmware file>\n");
    exit(1);
  }
  FILE *f = fopen(argv[2], "r");
  if (f == NULL) {
    perror("Could not open file.");
    return 1;
  }
  int numsectors = zaehleZeilen(f);
  fclose(f);
  f = fopen(argv[2], "r");
  if (f == NULL) {
    perror("Could not open file.");
    return 1;
  }
  char line[4500];
  int secno = 0;
  int serial = open_serial_port(argv[1], 115200);
  struct termios options;
  if (tcgetattr(serial, &options) < 0) {
    printf("Error getting attributes: %s\n", strerror(errno));
    return -1;
  }
  cfmakeraw(&options);
  cfsetspeed(&options, B115200);
  options.c_cflag &= ~CSTOPB;
  options.c_cflag |= CLOCAL;
  options.c_cflag |= CREAD;
  options.c_cflag |= CRTSCTS;
  options.c_cc[VTIME] = 5;
  options.c_cc[VMIN] = 1;
  if (tcsetattr(serial, TCSANOW, &options) < 0) {
    printf("Error setting attributes: %s\n", strerror(errno));
    return -1;
  }
  tcflush(serial, TCIFLUSH);
  sleep(2);
  int must_repeat = 0;
  int length;
  do {
    if (!must_repeat) {
      if (!fgets(line, sizeof(line), f)) {
        char s = 'Y';
        write(serial, &s, 1);
        if (tcdrain(serial) == -1) {
          perror("tcdrain");
          close(serial);
          return 1;
        }
        break;
      } else {
        if (line[0] == '#')
          continue;
        if (line[0] == ' ')
          continue;
        length = 4100;
        secno++;
        must_repeat = 0;
      }
    }
    sleep(1);
    char s = '\n';
    write(serial, &s, 1);
    if (tcdrain(serial) == -1) {
      perror("tcdrain");
      close(serial);
      return 1;
    }
    s = 'X';
    write(serial, &s, 1);
    if (tcdrain(serial) == -1) {
      perror("tcdrain");
      close(serial);
      return 1;
    }
    printf("Flashing Sector #%d/%d ...", secno, numsectors);
    fflush(stdout);
    {
      write(serial, &line[0], 4100);
      if (tcdrain(serial) == -1) {
        perror("tcdrain");
        close(serial);
        return 1;
      }
    }
    
    size_t received = 0;
    char received_char;
    while (received < 1) {
      struct timespec tim, tim2;
      tim.tv_sec = 0;
      tim.tv_nsec = 400000000;
      if (nanosleep(&tim, &tim2) < 0) {
        printf("Nano sleep system call failed \n");
        return -1;
      }
      ssize_t r = read(serial, &received_char, 1);
      if (r < 0) {
        perror("failed to read from port");
        return -1;
      }
      if (r == 0) {
        printf("Timeout\n");
        break;
      }
      received += r;
      if (received_char == 'A') {
        printf("OK.\n");
      }
      if (received_char == 'N') {
        printf("FAIL.\n");
        must_repeat = 1;
      }
    }
  } while (1);
  close(serial);
  fclose(f);
  return 0;
}
